from __future__ import annotations

import asyncio
from typing import Awaitable, Callable, Dict, Generator, TYPE_CHECKING, Sequence, Union

from ferris.client import Client
from .models import Command

if TYPE_CHECKING:
    from ferris.message import Message

    _BasePrefixT = Union[str, Sequence[str]]
    PrefixT = Union[
        _BasePrefixT,
        Callable[['Bot', Message], Union[_BasePrefixT, Awaitable[_BasePrefixT]]],
    ]


class CommandSink:
    """Represents a sink of commands.
    This will both mixin to :class:`.Bot` and :class:`.CommandGroup`.

    Attributes
    ----------
    mapping: Dict[str, :class:`.Command`]
        A full mapping of command names to commands.
    """

    def __init__(self) -> None:
        self.mapping: Dict[str, Command] = {}

    def walk_commands(self) -> Generator[Command]:
        """Returns a generator that walks through all of the commands
        this sink holds.

        Returns
        -------
        Generator[:class:`.Command`]
        """
        seen = set()
        for command in self.mapping.values():
            if command not in seen:
                seen.add(command)
                yield command

    @property
    def commands(self) -> None:
        """List[:class:`.Command`]: A list of the commands this sink holds."""
        return list(self.walk_commands())


class Bot(Client, CommandSink):
    """Represents a client connection to Discord with extra command handling support.

    Parameters
    ----------
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The event loop to use for the client. If not passed, then the default event loop is used.
    prefix
        The prefix the bot will listen for. This is required.
    max_messages_count: Optional[int]
        The maximum number of messages to store in the internal message buffer.
        Defaults to ``1000``.
    """

    def __init__(self, prefix: PrefixT, **kwargs) -> None:
        super().__init__(**kwargs)
        CommandSink.__init__(self)

        self.prefix: PrefixT = prefix



from __future__ import annotations

from typing import TypedDict, List, TYPE_CHECKING


if TYPE_CHECKING:
    from .base import Snowflake
    from .channel import ChannelPayload
    from .member import MemberPayload


__all__ = ('GuildPayload',)


class GuildPayload(TypedDict):
    id: Snowflake
    owner_id: Snowflake
    name: str
    channels: List[ChannelPayload]
    members: List[MemberPayload]



__author__ = 'Cryptex & jay3332'
__version__ = '0.0.b1'
# 0.1.0 for the finished release


import urllib.request

from .base import *
from .channel import *
from .client import *
from .connection import *
from .enums import *
from .errors import *
from .guild import *
from .http import *
from .invite import *
from .member import *
from .message import *
from .user import *
from .utils import *


def create_user(username: str, password: str, email: str) -> PartialUser:
    """
    Creates a new user.

    Parameters
    ----------
    username: str
        The user's username.
    password: str
        The user's password.
    email: str
        The user's email.

    Returns
    -------
    PartialUser
        The created user.
    """
    resp = urllib.request.urlopen(
        urllib.request.Request(
            f'{HTTPClient.API_BASE_URL}/users',
            data=to_json(
                {'email': email, 'password': password, 'username': username}
            ).encode('utf-8'),
            headers={'Content-Type': 'application/json'},
            method='POST',
        )
    )
    content = resp.read().decode('utf-8')
    js = from_json(content)

    return PartialUser(js)



from typing import TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from .base import Snowflake


class InvitePayload(TypedDict):
    code: str
    owner_id: Snowflake
    guild_id: Snowflake
    created_at: int
    uses: int
    max_uses: int
    max_age: int



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, List

from .base import BaseObject
from .message import Message
from .utils import sanitize_id

if TYPE_CHECKING:
    from .connection import Connection
    from .types import Data, Snowflake, Id
    from .types.channel import ChannelPayload
    from .guild import Guild

__all__ = ('Channel',)


class Channel(BaseObject):
    """
    Represents a channel in FerrisChat.
    """

    __slots__ = ('_connection', '_name', '_guild_id')

    def __init__(
        self, connection: Connection, data: Optional[ChannelPayload], /
    ) -> None:
        self._connection: Connection = connection
        self._process_data(data)

    def _process_data(self, data: Optional[ChannelPayload], /) -> None:
        if not data:
            return

        self._store_snowflake(data.get('id'))

        self._name: Optional[str] = data.get('name')

        self._guild_id: Optional[Snowflake] = data.get('guild_id')

    async def fetch_message(self, message_id: Id) -> Message:
        """
        |coro|

        Fetches a message from this channel.

        Parameters
        ----------
        message_id: int
            The ID of the message to fetch.

        Returns
        -------
        Message
        """
        message_id = sanitize_id(message_id)
        m = await self._connection.api.messages(message_id).get()
        return Message(self._connection, m)

    async def send(self, content: str) -> Message:
        """
        |coro|

        Sends a message to this channel.

        Parameters
        ----------
        content: str
            The content of the message.

        Returns
        -------
        Message
            The message that was sent.
        """
        m = await self._connection.api.channels(self.id).messages.post(
            json={'content': content}
        )
        return Message(self._connection, m)

    async def fetch_messages(self, limit: int = 100) -> List[Message]:
        """
        |coro|

        Fetches messages from this channel.

        Parameters
        ----------
        limit: int
            The maximum number of messages to fetch.
            Defaults to 100.
            Set it to 9223372036854775807 to fetch all messages.

        Returns
        -------
        List[Message]
        """
        data = await self._connection.api.channels(self.id).messages.get(limit=limit)
        return [Message(self._connection, m) for m in data]

    async def edit(self, name: str) -> Channel:
        """
        |coro|

        Edits this channel.

        Parameters
        ----------
        name: str
            The new name of the channel.

        Returns
        -------
        Channel
            The channel that was edited.
        """
        data = {'name': name}
        c = await self._connection.api.channels(self.id).patch(json=data)
        self._process_data(c)
        return self

    async def delete(self) -> None:
        """
        |coro|

        Deletes this channel.
        """
        await self._connection.api.channels(self.id).delete()

    @property
    def guild(self, /) -> Optional[Guild]:
        """Guild: The guild that this channel belongs to."""
        return self._connection.get_guild(self._guild_id)

    @property
    def guild_id(self) -> Optional[Snowflake]:
        """int: The ID of the guild this channel belongs to."""
        return self._guild_id

    @property
    def name(self, /) -> Optional[str]:
        """str: The name of this channel."""
        return self._name

    def __repr__(self, /) -> str:
        return f'<Channel id={self.id} name={self.name} guild_id={self.guild_id}>'

    def __del__(self, /) -> None:
        if not hasattr(self, '_connection'):
            return

        self._connection.deref_channel(self.id)



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Awaitable, ClassVar, Dict, Optional
from urllib.parse import quote

import aiohttp

from . import __version__
from .errors import (
    BadRequest,
    FerrisServerError,
    FerrisUnavailable,
    Forbidden,
    HTTPException,
    MissingImplementation,
    NotFound,
    Unauthorized,
)
from .utils import from_json

if TYPE_CHECKING:
    from .types import Data, SupportsStr

log = logging.getLogger(__name__)

__all__ = (
    'APIRouter',
    'HTTPClient',
)


class APIRouter:
    __slots__ = ('__current_route', '__http_client')

    def __init__(self, http: HTTPClient, route: str = '', /) -> None:
        self.__current_route: str = route
        self.__http_client: HTTPClient = http

    @property
    def url(self, /) -> str:
        return HTTPClient.API_BASE_URL + self.__current_route

    def _make_new(self, route: str, /) -> APIRouter:
        return self.__class__(self.__http_client, route)

    def __getattr__(self, route: str, /) -> APIRouter:
        return self._make_new(f"{self.__current_route}/{route}")

    def __call__(self, route: Optional[SupportsStr], /) -> APIRouter:
        return self._make_new(f"{self.__current_route}/{quote(str(route))}")

    def request(self, method, /, **kwargs) -> Awaitable[Optional[Data]]:
        return self.__http_client.request(self.url, method, **kwargs)

    def get(self, /, **kwargs) -> Awaitable[Optional[Data]]:
        return self.request('GET', **kwargs)

    def post(self, /, **kwargs) -> Awaitable[Optional[Data]]:
        return self.request('POST', **kwargs)

    def put(self, /, **kwargs) -> Awaitable[Optional[Data]]:
        return self.request('PUT', **kwargs)

    def delete(self, /, **kwargs) -> Awaitable[Optional[Data]]:
        return self.request('DELETE', **kwargs)

    def patch(self, /, **kwargs) -> Awaitable[Optional[Data]]:
        return self.request('PATCH', **kwargs)


class HTTPClient:
    API_BASE_URL: ClassVar[str] = 'https://api.ferris.chat/v0'

    MAX_TRIES: ClassVar[int] = 3
    USER_AGENT: ClassVar[
        str
    ] = f"FerrisWheel (https://github.com/Cryptex-github/ferriswheel v{__version__})"

    __slots__ = ('__token', '__session', '_buckets_lock', '_api_router')

    def __init__(self, token: str, /) -> None:
        self.__token: str = token
        self.__session: aiohttp.ClientSession = aiohttp.ClientSession(
            headers={"User-Agent": self.USER_AGENT, "Authorization": self.__token}
        )

        self._buckets_lock: Dict[str, asyncio.Event] = {}
        self._api_router: APIRouter = APIRouter(self)

    @property
    def token(self) -> str:
        return self.__token

    @property
    def api(self) -> APIRouter:
        return self._api_router

    @property
    def session(self) -> aiohttp.ClientSession:
        return self.__session

    @classmethod
    async def from_email_and_password(cls, email: str, password: str) -> HTTPClient:
        log.info("Retriving token from email and password")
        for tries in range(cls.MAX_TRIES):
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    f"{cls.API_BASE_URL}/auth",
                    headers={"Email": email, "Password": password},
                ) as response:
                    content = await response.text()

                    if 400 > response.status >= 200:
                        token = from_json(content)['token']
                        return cls(token)
                        log.info("Successfully Retrived token")

                    if response.status == 400:
                        data = from_json(content)
                        reason = data.get('reason')
                        location = data.get('location', {})
                        line = location.get('line')
                        character = location.get('character')

                        raise BadRequest(
                            response, f"{reason}\nLine: {line} Character: {character}"
                        )

                    if response.status == 404:
                        raise NotFound(response, content)

                    if response.status == 401:
                        raise Unauthorized(response, content)

                    if response.status == 403:
                        raise Forbidden(response, content)

                    if 500 <= response.status < 600:
                        if tries == 1:
                            try:
                                data = from_json(content)
                                reason = data.get('reason')
                            except:  # TODO: Fix broad except
                                reason = content

                            raise FerrisUnavailable(response, reason)

                        continue

        raise HTTPException(response, content)

    async def request(self, url: str, method: str, /, **kwargs) -> Optional[Data]:
        bucket_key = f"{method} {url}"
        bucket = self._buckets_lock.get(bucket_key)
        if bucket is None:
            self._buckets_lock[bucket_key] = bucket = asyncio.Event()
            bucket.set()

        headers = {}

        if "data" in kwargs:
            headers["Content-Type"] = "application/json"

        if not bucket.is_set():
            await bucket.wait()

        for tries in range(self.MAX_TRIES):
            async with self.__session.request(
                method, url, headers=headers, **kwargs
            ) as response:
                content = await response.text()

                log.debug(f"{method} {url} Returned {response.status} with {content}")

                if 400 > response.status >= 200:
                    return from_json(content)

                if response.status == 429:
                    data = from_json(content)
                    sleep = data.get("retry_after", 0)
                    log.warning(
                        f"We have been ratelimited on {method} {url}, retrying in {sleep} seconds"
                    )
                    bucket.clear()
                    await asyncio.sleep(sleep)
                    bucket.set()
                    continue

                if response.status == 400:
                    data = from_json(content)
                    reason = data.get('reason')
                    location = data.get('location', {})
                    line = location.get('line')
                    character = location.get('character')
                    raise BadRequest(
                        response, f"{reason}\nLine: {line} Character: {character}"
                    )

                if response.status == 404:
                    raise NotFound(response, content)

                if response.status == 401:
                    raise Unauthorized(response, content)

                if response.status == 403:
                    raise Forbidden(response, content)

                if 500 <= response.status < 600:
                    if tries == 1:
                        if response.status == 500:
                            try:
                                data = from_json(content)
                                reason = data.get('reason')
                            except:  # TODO: Fix broad except
                                reason = content

                            raise FerrisServerError(response, reason)

                        if response.status == 501:
                            raise MissingImplementation(response, content)

                        raise FerrisUnavailable(response, content)

                    continue

        raise HTTPException(response, content)



from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from typing import (
    Any,
    overload,
    Literal,
    TypeVar,
    TYPE_CHECKING,
    Optional,
)

from .utils import get_snowflake_creation_date


E = TypeVar('E', bound='BaseObject')

__all__ = ('SnowflakeObject', 'BaseObject', 'Object')

if TYPE_CHECKING:
    from .types import Data, Snowflake


class SnowflakeObject(ABC):
    """An abstract base class representing objects that have a snowflake ID."""

    __slots__ = ('_id',)

    def __init__(self, /) -> None:
        self._id: Optional[int] = None

    def _store_snowflake(self, id: Optional[Snowflake], /) -> None:
        self._id = id

    @property
    def id(self, /) -> Optional[int]:
        """int: The snowflake ID of this object."""
        try:
            return self._id  # type: ignore
        except AttributeError:
            return None


class BaseObject(SnowflakeObject, ABC):
    """The base class that all FerrisChat-related objects will inherit from."""

    __slots__ = ()

    @property
    def created_at(self, /) -> datetime:
        """:class:`datetime.datetime`: The creation timestamp of this object."""
        return get_snowflake_creation_date(self.id)

    @abstractmethod
    def _process_data(self, data: Any, /) -> None:
        raise NotImplementedError

    def __hash__(self, /) -> int:
        return hash(self.id)

    @overload
    def __eq__(self: E, other: E, /) -> bool:  # type: ignore
        ...

    @overload
    def __eq__(self, other: Any, /) -> Literal[False]:
        ...

    def __eq__(self, other: Any, /) -> bool:
        return isinstance(other, self.__class__) and other.id == self.id

    def __ne__(self, other: Any, /) -> bool:
        return not self.__eq__(other)


class Object(SnowflakeObject):
    """Represents an anonymous FerrisChat-related object that has a snowflake ID.
    Instances of this may be constructed by yourself, safely.

    Parameters
    ----------
    id: int
        The snowflake ID of this object.
    """

    __slots__ = ()

    def __init__(self, id: int, /) -> None:
        super().__init__()
        self._store_snowflake(id)



from datetime import datetime
from typing import Any, Callable, Iterable, Optional, TypeVar

from .types import Id, Snowflake

__all__ = (
    'to_json',
    'from_json',
    'get_snowflake_creation_date',
    'find',
    'dt_to_snowflake',
)

T = TypeVar('T', covariant=True)

try:
    import orjson

    HAS_ORJSON = True
except ImportError:
    import json

    HAS_ORJSON = False

FERRIS_EPOCH_MS: int = 1_577_836_800_000


FERRIS_EPOCH: int = 1_577_836_800


if HAS_ORJSON:

    def to_json(obj: Any) -> str:
        return orjson.dumps(obj).decode('utf-8')

    def from_json(json_str: str) -> Any:
        if not json_str:
            return None
        return orjson.loads(json_str)


else:

    def to_json(obj: Any) -> str:
        return json.dumps(obj, ensure_ascii=True)

    def from_json(json_str: str) -> Any:
        if not json_str:
            return None
        return json.loads(json_str)


def sanitize_id(id: Id = None) -> Snowflake:
    """Sanitizes an ID.

    Parameters
    ----------
    id: Id
        The ID to sanitize.

    Returns
    -------
    Snowflake
        The sanitized ID.
    """
    return getattr(id, 'id', id) if id else None


def get_snowflake_creation_date(snowflake: int) -> datetime:
    """Returns the creation timestamp of the given snowflake.

    Parameters
    ----------
    snowflake: int
        The snowflake to get the creation timestamp of.

    Returns
    -------
    :class:`datetime.datetime`
        The creation date of the snowflake.
    """
    seconds = ((snowflake >> 64) + FERRIS_EPOCH_MS) / 1000
    return datetime.utcfromtimestamp(seconds)


def dt_to_snowflake(dt: datetime) -> int:
    """Generates a random snowflake that was created on the given datetime.

    Parameters
    ----------
    dt: :class:`datetime.datetime`
        The target creation date of the snowflake to generate.

    Returns
    -------
    int
        The generated snowflake.
    """
    timestamp = dt.timestamp() * 1000 - FERRIS_EPOCH_MS
    return int(timestamp) << 64


def find(predicate: Callable[[T], Any], iterable: Iterable[T]) -> Optional[T]:
    """Finds the first element in the iterable that satisfies the predicate.

    Parameters
    ----------
    predicate: Callable[[Any], Any]
        The predicate to check each element against.
    Iterable: Iterable[Any]
        The iterable to search.

    Returns
    -------
    Any
        The first element in the iterable that satisfies the predicate.
        Can be None if no element satisfies the predicate.

    """
    for element in iterable:
        if predicate(element):
            return element
    return None



from __future__ import annotations

import logging
from types import coroutine
from typing import TYPE_CHECKING, Coroutine, Set, Union

import aiohttp

from .errors import Reconnect, WebsocketException
from .handler import EventHandler
from .utils import from_json, to_json

if TYPE_CHECKING:
    from .client import Client
    from .http import HTTPClient
    from .types import Data
    from .types.ws import WsConnectionInfo


log = logging.getLogger(__name__)


class Websocket:
    """The class that interfaces with FerrisChat's websockets."""

    def __init__(self, client: Client) -> None:
        self.ws: aiohttp.ClientWebSocketResponse

        self._handler: EventHandler = EventHandler(client._connection)

        self._http: HTTPClient = client._connection._http
        self.dispatch: Coroutine = client.dispatch
        self._ws_url: str = ''

    async def prepare(self) -> None:
        """Retrieves the URL needed for websocket connection."""
        response: WsConnectionInfo = await self._http.api.ws.info.get()  # type: ignore
        self._ws_url = response['url']

    def handle(self, data: dict) -> None:
        """Handles a message received from the websocket."""
        log.debug(f'Received: {data}')
        self._handler.handle(data)

    def _parse_and_handle(self, data: Union[str, bytes]) -> None:
        if isinstance(data, (str, bytes)):
            _data: dict = from_json(data)
            self.handle(_data)

    async def send(self, data: dict, /):
        _data = to_json(data)
        await self.ws.send_str(_data)

    async def connect(self) -> None:
        """Establishes a websocket connection with FerrisChat."""
        if not self._ws_url:
            await self.prepare()

        self.ws = await self._http.session.ws_connect(self._ws_url, heartbeat=45)

        await self.send(
            {'c': 'Identify', 'd': {'token': self._http.token, 'intents': 0}}
        )

        self.dispatch('connect')
        async for message in self.ws:
            if message.type in {aiohttp.WSMsgType.TEXT, aiohttp.WSMsgType.BINARY}:
                self._parse_and_handle(message.data)
            elif message.type is aiohttp.WSMsgType.ERROR:
                log.error(f'Websocket error: {message.data}')
                raise WebsocketException(message.data)
            elif message.type in (
                aiohttp.WSMsgType.CLOSED,
                aiohttp.WSMsgType.CLOSING,
                aiohttp.WSMsgType.CLOSE,
            ):
                log.info('Websocket closed, attempting to reconnect.')
                raise Reconnect

    async def close(self, code) -> None:
        """Closes the current websocket connection."""
        if ws := getattr(self, 'ws', None):
            if not ws.closed:
                await self.ws.close(code=code)



from typing import Optional

from aiohttp import ClientResponse

__all__ = (
    'FerrisException',
    'HTTPException',
    'BadRequest',
    'Unauthorized',
    'Forbidden',
    'NotFound',
    'FerrisServerError',
    'FerrisUnavailable',
    'WebsocketException',
    'MissingImplementation',
    'Reconnect',
)


class FerrisException(Exception):
    """Base class for all ferriswheel-originated exceptions.
    Theoretically, you could use this exception to catch all ferris exceptions.
    """

    pass


class HTTPException(FerrisException):
    """
    Base class for all ferris HTTP exceptions.

    Attributes
    ----------
    status: int
        The status code of the response.
    resp: :class:`aiohttp.ClientResponse`
        The aiohttp response object.
    """

    def __init__(self, resp: ClientResponse, content: Optional[str] = None):
        content = content or resp.reason
        self.status = resp.status
        self.resp = resp
        super().__init__(content)


class BadRequest(HTTPException):
    """The request was invalid or cannot be otherwise served."""

    pass


class Unauthorized(HTTPException):
    """The request requires user authentication."""

    pass


class Forbidden(HTTPException):
    """The request was a valid request,
    but the server is refusing to respond to it, as you don't have enough permissions to do that.
    """

    pass


class NotFound(HTTPException):
    """The server cannot find what you requesed for."""

    pass


class MissingImplementation(HTTPException):
    """You just discovered Ferris at work! Don't tell anybody, but this is planned for the future!"""

    pass


class FerrisServerError(HTTPException):
    """Ferris is currently unavailable. Looks like a JS rabbit got into the server."""

    pass


class FerrisUnavailable(HTTPException):
    """SSHHH! Ferris is sleeping, and can't answer the request right now.
    Try again later.
    """

    pass


class WebsocketException(FerrisException):
    """Base class for all websocket exceptions."""

    pass


class Reconnect(WebsocketException):
    """Signal to reconnect to the websocket."""

    pass



from __future__ import annotations

from typing import TypedDict, TYPE_CHECKING


if TYPE_CHECKING:
    from .base import Snowflake
    from .user import UserPayload
    from .guild import GuildPayload

__all__ = ('MemberPayload',)


class MemberPayload(TypedDict):
    user_id: Snowflake
    user: UserPayload
    guild: GuildPayload
    guild_id: Snowflake



from __future__ import annotations
from ferris.types.base import Snowflake

from typing import TYPE_CHECKING, Dict, List, Optional

from .base import BaseObject
from .guild import Guild

if TYPE_CHECKING:
    from .connection import Connection
    from .types import Data
    from .types.user import UserPayload

__all__ = ('PartialUser', 'User')


class PartialUser(BaseObject):
    """
    Represents a Partial FerrisChat User.
    """

    __slots__ = ('_name',)

    def __init__(self, data: Optional[UserPayload], /) -> None:
        self._process_data(data)

    def _process_data(self, data: Optional[UserPayload]) -> None:
        if not data:
            return

        self._store_snowflake(data.get('id'))

        self._name: Optional[str] = data.get('name')

    @property
    def name(self) -> Optional[str]:
        """str: The user's name."""
        return self._name

    def __repr__(self) -> str:
        return f'<PartialUser id={self.id} name={self.name!r}>'


class User(BaseObject):
    """
    Represents a FerrisChat user.
    """

    __slots__ = ('_connection', '_name', '_guilds', '_avatar')

    def __init__(self, connection: Connection, data: UserPayload, /) -> None:
        self._connection: Connection = connection
        self._process_data(data)

    def _process_data(self, data: UserPayload, /) -> None:
        self._store_snowflake(data.get('id'))

        self._name: Optional[str] = data.get('name')

        self._avatar: Optional[str] = data.get('avatar')

        self._guilds: Dict[int, Guild] = {}

        for g in data.get('guilds') or []:
            guild_id = g.get('id')

            if guild_id in self._guilds:
                guild = self._connection.get_guild(guild_id)
                guild._process_data(g)

            if guild_id not in self._guilds:
                guild = Guild(self._connection, g)
                self._connection.store_guild(guild)

            self._guilds[guild_id] = self._connection.get_guild(guild_id)

        # self._flags = data.get('flags')
        # UserFlag after ferrischat implemented it

    async def edit(
        self,
        username: Optional[str] = None,
        email: Optional[str] = None,
        password: Optional[str] = None,
    ) -> User:
        """|coro|

        Edits the user.

        Parameters
        ----------
        username : Optional[str]
            The new username.
        email : Optional[str]
            The new email.
        password : Optional[str]
            The new password.

        Returns
        -------
        User
            The edited user.
        """
        payload = {
            'username': username,
            'email': email,
            'password': password,
        }
        user = await self._connection.api.users(self.id).patch(payload)
        self._process_data(user)
        return self

    async def fetch_guilds(self) -> List[Guild]:
        """|coro|

        Fetches all the guilds this user is in.

        Returns
        -------
        List[:class:`~.Guild`]
            A list of the guilds this user is in.

        .. warning::
            This method will do nothing as FerrisChat has not implemented this feature yet.
        """
        ...

    @property
    def name(self, /) -> Optional[str]:
        """str: The username of this user."""
        return self._name

    @property
    def guilds(self, /) -> List[Guild]:
        """List[:class:`~.Guild`]: A list of the guilds this user is in."""
        return list(self._guilds.values())

    @property
    def avatar(self, /) -> Optional[str]:
        """str: The avatar url of this user."""
        return self._avatar

    def __del__(self, /) -> None:
        if not hasattr(self, '_connection'):
            return

        self._connection.deref_user(self.id)

    def __repr__(self, /) -> str:
        return f'<User id={self.id} name={self.name!r}>'



from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict, List

if TYPE_CHECKING:
    from .base import Snowflake
    from .guild import GuildPayload


__all__ = ('UserPayload',)


class UserPayload(TypedDict):
    id: Snowflake
    name: str
    guilds: List[GuildPayload]
    flags: int
    avatar: str



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from .base import BaseObject
from .user import User

if TYPE_CHECKING:
    from .guild import Guild
    from .connection import Connection
    from .types import Data
    from .types.member import MemberPayload

__all__ = ('Member',)


class Member(BaseObject):
    """Represents a member of a :class:`~.Guild`."""

    __slots__ = ('_connection', '_user', '_guild', '_guild_id')

    def __init__(self, connection: Connection, data: MemberPayload, /) -> None:
        self._connection: Connection = connection
        self._process_data(data)

    def _process_data(self, data: Optional[MemberPayload], /) -> None:
        if not data:
            return

        self._store_snowflake(data.get('user_id'))

        if user := self._connection.get_user(self.id):
            self._user = user
            self._user._process_data(data.get('user') or {})
        else:
            self._user: Optional[User] = User(self._connection, data.get('user') or {})  # type: ignore
            self._connection.store_user(self._user)

        self._guild_id: Optional[int] = data.get('guild_id')

        if guild := self._connection.get_guild(self._guild_id):
            guild._process_data(data.get('guild') or {})
        else:

            from .guild import Guild

            guild: Optional[Guild] = Guild(
                self._connection, data.get('guild', {}) or {}
            )

            self._connection.store_guild(guild)

        self._guild: Optional[Guild] = guild

    async def edit(self) -> None:
        """|coro|

        Edits this member.

        .. warning::
            This method will do nothing as FerrisChat has not implemented this feature yet.
        """
        ...

    async def delete(self) -> None:
        """|coro|

        Kicks this member from it's guild.

        .. warning::
            This method will do nothing as FerrisChat has not implemented this feature yet.
        """
        ...

    @property
    def user(self, /) -> Optional[User]:
        """:class:`~.User`: The user that belongs to this member."""
        return self._user

    @property
    def guild(self, /) -> Optional[Guild]:
        """:class:`~.Guild`: The guild that this member belongs to."""
        return self._guild

    @property
    def guild_id(self, /) -> Optional[int]:
        """int: The ID of the guild that this member belongs to."""
        return self._guild_id

    def __repr__(self, /) -> str:
        return f'<Member id={self.id} user={self.user!r} guild={self.guild!r}>'



from typing import Protocol, Union, runtime_checkable, Optional

__all__ = ('SupportsStr', 'SupportsId', 'Id', 'Snowflake')


class SupportsStr(Protocol):
    def __str__(self) -> str:
        ...


Snowflake = int


@runtime_checkable
class SupportsId(Protocol):
    __slots__ = ()

    id: Snowflake


Id = Optional[Union[SupportsId, Snowflake]]



from __future__ import annotations

import asyncio
import logging
import signal
from collections import defaultdict
from typing import (TYPE_CHECKING, Awaitable, Callable, Dict, List, Optional,
                    overload)

from ferris.types import message

from .channel import Channel
from .connection import Connection
from .errors import Reconnect, WebsocketException
from .guild import Guild
from .invite import Invite
from .message import Message
from .user import User
from .utils import sanitize_id
from .websocket import Websocket

if TYPE_CHECKING:
    from .types import Id

log = logging.getLogger(__name__)

__all__ = ('Dispatcher', 'Client')

# https://github.com/Rapptz/discord.py/blob/main/discord/client.py#L81
# https://github.com/python/cpython/blob/main/Lib/asyncio/runners.py

def _cancel_tasks(loop: asyncio.AbstractEventLoop) -> None:
    tasks = {t for t in asyncio.all_tasks(loop=loop) if not t.done()}

    if not tasks:
        return
    
    log.info(f'Cancelling {len(tasks)} tasks.')

    for task in tasks:
        task.cancel()

    loop.run_until_complete(asyncio.gather(*tasks, return_exceptions=True))
    log.info('Cancelled all tasks.')

    for task in tasks:
        if task.cancelled():
            continue
        if task.exception() is not None:
            loop.call_exception_handler({
                'message': 'Unhandled exception during Client.run shutdown.',
                'exception': task.exception(),
                'task': task
            })

def _cleanup_loop(loop: asyncio.AbstractEventLoop) -> None:
    try:
        _cancel_tasks(loop)
        loop.run_until_complete(loop.shutdown_asyncgens())
    finally:
        log.info('Closing the event loop.')
        loop.close()


class Dispatcher:
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop: asyncio.AbstractEventLoop = loop
        self.event_handlers: defaultdict[
            str, List[Callable[..., Awaitable]]
        ] = defaultdict(list)

    def dispatch(self, event: str, *args, **kwargs) -> asyncio.Future:
        coros = []

        if callback := getattr(self, f'on_{event}', False):
            coros.append(callback(*args, **kwargs))

        if callbacks := self.event_handlers.get(event):
            coros += [cb(*args, **kwargs) for cb in callbacks]

        return asyncio.ensure_future(asyncio.gather(*coros))

    def add_listener(self, event: str, callback: Callable[..., Awaitable]) -> None:
        if event.startswith('on_'):
            event = event[3:]

        self.event_handlers[event].append(callback)

    def remove_listener(self, event: str, callback: Callable[..., Awaitable]) -> None:
        if event.startswith('on_'):
            event = event[3:]

        self.event_handlers[event].remove(callback)

    def clear_listeners(self) -> None:
        self.event_handlers.clear()

    def stop_listening_to(self, event: str, /) -> None:
        if event.startswith('on_'):
            event = event[3:]

        self.event_handlers[event].clear()
        del self.event_handlers[event]

    def listen(self, event: Optional[str] = None):
        def decorator(func: Callable[..., Awaitable]) -> Callable[..., Awaitable]:
            store = event

            if store is None:
                store = func.__name__

            if store.startswith('on_'):
                store = store[3:]

            self.add_listener(store, func)
            return func

        return decorator

    def event(self, func: Callable[..., Awaitable]) -> Callable[..., Awaitable]:
        event = func.__name__

        if not event.startswith('on_'):
            event = 'on_' + event

        setattr(self, event, func)

        return func


class EventTemplateMixin:
    async def on_login(self) -> None:
        """|coro|

        Called when the client has logged in.
        """
        pass

    async def on_close(self) -> None:
        """|coro|

        Called when the client is closing.
        This could be used for clean-up/memory freeing.
        """
        pass

    async def on_connect(self) -> None:
        """|coro|
        Called when the client has connected to FerrisChat ws.
        """
        pass


class Client(Dispatcher, EventTemplateMixin):
    """Represents a client connection to FerrisChat.

    Parameters
    ----------
    loop: Optional[:class:`asyncio.AbstractEventLoop`]
        The event loop to use for the client. If not passed, then the default event loop is used.
    max_messages_count: Optional[int]
        The maximum number of messages to store in the internal message buffer.
        Defaults to ``1000``.
    """

    def __init__(
        self, /, loop: Optional[asyncio.AbstractEventLoop] = None, **options
    ) -> None:
        self.loop = loop or asyncio.get_event_loop()
        self._is_closed: bool = False
        self._connection: Connection = Connection(self.loop, self.dispatch, **options)
        self._is_ready = self._connection._is_ready
        super().__init__(self.loop)

    @property
    def user(self) -> Optional[User]:
        """Returns the client's user."""
        return self._connection.user

    @property
    def is_ready(self) -> bool:
        """Returns whether the client is ready to use."""
        return self._is_ready.done()

    @property
    def is_closed(self) -> bool:
        """Returns whether the client is closed."""
        return self._is_closed

    async def wait_until_ready(self) -> None:
        """|coro|

        Waits until the client is ready to use.
        """
        return await self._is_ready

    @property
    def guilds(self) -> List[Guild]:
        """Returns a list of the client's guilds."""
        return list(self._connection._guilds.values())

    def _initialize_connection(self, token: str, /) -> None:
        self._connection._initialize_http(token)

    async def join_guild(self, code: str) -> None:
        """|coro|

        Joins a guild with an invite code.

        Parameters
        ----------
        code: str
            The invite code to use.
        """
        await self._connection.api.invites(code).post()

    async def fetch_invite(self, code: str) -> None:
        """|coro|

        Fetches an invite by code.

        Parameters
        ----------
        code: str
            The invite code to fetch.
        """
        i = await self._connection.api.invites(code).get()
        return Invite(self._connection, i)

    async def create_guild(self, name: str) -> Guild:
        """|coro|

        Creates a new guild.

        Parameters
        ----------
        name: str
            The name of the guild.

        Returns
        -------
        :class:`Guild`
            The guild created.
        """
        g = await self._connection.api.guilds.post(json={'name': name})
        return Guild(self._connection, g)

    def get_message(self, id: Id) -> Optional[Message]:
        """
        Gets a message from the internal message buffer.

        Parameters
        ----------
        id: int
            The ID of the message to get.

        Returns
        -------
        Optional[:class:`Message`]
            The message with the given ID, or ``None`` if it does not exist.
        """
        id = sanitize_id(id)
        return self._connection.get_message(id)

    def get_channel(self, id: Id) -> Optional[Channel]:
        """
        Gets a channel from the internal channel buffer.

        Parameters
        ----------
        id: int
            The ID of the channel to get.

        Returns
        -------
        Optional[:class:`Channel`]
            The channel with the given ID, or ``None`` if it does not exist.
        """
        id = sanitize_id(id)
        return self._connection.get_channel(id)

    def get_user(self, id: Id) -> Optional[User]:
        """
        Gets a user from the internal user buffer.

        Parameters
        ----------
        id: int
            The ID of the user to get.

        Returns
        -------
        Optional[:class:`User`]
            The user with the given ID, or ``None`` if it does not exist.
        """
        id = sanitize_id(id)
        return self._connection.get_user(id)

    async def fetch_message(self, id: Id) -> Message:
        """|coro|

        Fetches a message from the internal message buffer.

        Parameters
        ----------
        id: int
            The ID of the message to fetch.

        Returns
        -------
        :class:`Message`
            The message with the given ID.
        """
        id = sanitize_id(id)
        m = await self._connection.api.messages(id).get()
        return Message(self._connection, m)

    async def fetch_channel(self, id: Id) -> Channel:
        """|coro|

        Fetches a channel by ID.

        Parameters
        ----------
        id: int
            The ID of the channel to fetch.

        Returns
        -------
        :class:`Channel`
            The channel with the given ID.
        """
        id = sanitize_id(id)
        c = await self._connection.api.channels(id).get()
        return Channel(self._connection, c)

    async def fetch_user(self, id: Id) -> User:
        """|coro|

        Fetches a user by ID.

        Parameters
        ----------
        id: int
            The ID of the user to fetch.

        Returns
        -------
        :class:`User`
            The user with the given ID.
        """
        id = sanitize_id(id)
        u = await self._connection.api.users(id).get()
        return User(self._connection, u)

    async def fetch_guild(
        self, id: Id, fetch_members: bool = False, fetch_channels: bool = True
    ) -> Guild:
        """|coro|

        Fetches a guild by ID.

        Parameters
        ----------
        id: int
            The guild's ID.
        fetch_members: Optional[bool], default False
            Whether to fetch the guild's members. Defaults to ``False``.
        fetch_channels: Optional[bool], default True
            Whether to fetch the guild's channels. Defaults to ``True``.

        Returns
        -------
        :class:`Guild`
            The guild fetched.

        Raises
        ------
        :exc:`NotFound`
            A guild with the given ID was not found.
        """
        id = sanitize_id(id)
        g = await self._connection.api.guilds(id).get(
            params={
                'members': str(fetch_members).lower(),
                'channels': str(fetch_channels).lower(),
            }
        )
        return Guild(self._connection, g)

    async def close(self) -> None:
        if self.is_closed:
            return
        self._is_closed = True

        if hasattr(self, 'ws'):
            await self.ws.close(code=1000)
        
        if http := getattr(self._connection, '_http', None):
            if session := getattr(http, 'session', None):
                await session.close()

        self.dispatch('close')


    stop = close

    @overload
    async def start(self, *, token: str) -> None:
        ...

    @overload
    async def start(self, *, email: str, passsword: str, id: Id) -> None:
        ...

    async def start(
        self,
        *,
        token: Optional[str] = None,
        email: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        """|coro|

        Establishes a websocket connection with FerrisChat's gateway.
        A form of authentication (email and password, or a token) must be provided.

        The `id` kwarg is only needed when an email and password is used to login.

        Parameters
        ----------
        token: Optional[str]
            The token used to authenticate.
        email: Optional[str]
            The email used to authenticate.
        password: Optional[str]
            The password used to authenticate.
        id: Optional[int]
            The ID of the guild to join.
        """

        if email is not None and password is not None and token is not None:
            raise ValueError('Cannot pass both email and password and token')

        if token is not None and (email is not None or password is not None):
            raise ValueError('Cannot pass both token and email or password')

        if token is None and (email is None or password is None):
            raise ValueError('Must pass either token or email and password')

        if token is not None:
            log.info("Logging in with Token")
            self._initialize_connection(token)
        else:
            log.info("Logging in with Email and Password")
            await self._connection._initialize_http_with_email(email, password)  # type: ignore

        self.dispatch('login')

        self.ws = Websocket(self)

        while not self.is_closed:
            try:
                await self.ws.connect()
            except Reconnect:
                del self.ws
                self.ws = Websocket(self)
                continue
            except WebsocketException:
                raise

    def run(self, *args, **kwargs):
        """A helper function equivalent to

        .. code-block:: python3

            try:
                loop.run_until_complete(start(*args, **kwargs))
            except KeyboardInterrupt:
                loop.run_until_complete(_cancel_all_tasks())
            finally:
                loop.close()

        If you want finer control over the event loop, use :meth:`Client.start` instead.
        """
        # https://github.com/Rapptz/discord.py/blob/master/discord/client.py#L608 and
        # https://github.com/python/cpython/blob/main/Lib/asyncio/runners.py

        loop = self.loop

        try:
            loop.add_signal_handler(signal.SIGINT, lambda: loop.stop())
            loop.add_signal_handler(signal.SIGTERM, lambda: loop.stop())
        except NotImplementedError:
            pass

        async def runner():
            try:
                await self.start(*args, **kwargs)
            finally:
                if not self.is_closed:
                    await self.close()

        def stop_loop_on_completion(f):
            loop.stop()

        future = asyncio.ensure_future(runner(), loop=loop)
        future.add_done_callback(stop_loop_on_completion)
        try:
            loop.run_forever()
        except KeyboardInterrupt:
            log.info('Received signal to terminate bot and event loop')
        finally:
            future.remove_done_callback(stop_loop_on_completion)
            log.info('Cleaning up tasks.')
            _cleanup_loop(loop)

        if not future.cancelled():
            try:
                return future.result()
            except KeyboardInterrupt:
                return None


from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional, cast

from .base import BaseObject
from .utils import sanitize_id
from .channel import Channel
from ferris.invite import Invite


if TYPE_CHECKING:
    from .member import Member
    from .connection import Connection
    from .types import Data, Id, Snowflake
    from .types import GuildPayload, ChannelPayload


__all__ = ('Guild',)


class Guild(BaseObject):
    """Represents a FerrisChat guild."""

    __slots__ = ('_connection', '_owner_id', '_name', '_channels', '_members')

    def __init__(self, connection: Connection, data: Optional[GuildPayload], /) -> None:
        self._connection: Connection = connection
        self._process_data(data)

    def _process_data(self, data: Optional[GuildPayload], /) -> None:
        if not data:
            return

        from .member import Member

        self._store_snowflake(data.get('id'))

        self._owner_id: Optional[int] = data.get('owner_id')
        self._name: Optional[str] = data.get('name')

        self._channels: Dict[int, Channel] = {}

        for c in data.get('channels') or []:
            if channel := self._connection.get_channel(c.get('id')):
                channel._process_data(c)
            else:
                channel = Channel(self._connection, c)
            self._channels[channel.id] = channel

        self._members: Dict[int, Member] = {}

        for m in data.get('members') or []:
            member = Member(self._connection, m)
            self._members[member.id] = member

    async def fetch_member(self, id: Id) -> Member:
        """|coro|

        Fetches a member from this guild.

        .. warning::
            This method will do nothing as FerrisChat has not implemented this feature yet.
        """
        ...

    async def fetch_invites(self) -> List[Invite]:
        """|coro|

        Fetches all the invites for this guild.

        Returns
        -------
        List[:class:`~.Invite`]
        """
        invites = await self._connection.api.guilds(self.id).invites.get()
        return [Invite(self._connection, i) for i in invites]

    async def create_invite(self, max_age: int = None, max_uses: int = None) -> Invite:
        """|coro|

        Creates an invite for this guild.

        Parameters
        ----------
        max_age: Optional[int]
            The maximum age of the invite, in seconds.
        max_uses: Optional[int]
            The maximum uses of the invite.

        Returns
        -------
        :class:`~.Invite`
        """
        payload = {
            'max_age': max_age,
            'max_uses': max_uses,
        }
        invite = await self._connection.api.guilds(self.id).invites.post(json=payload)
        return Invite(self._connection, invite)

    async def create_channel(self, name: str) -> Channel:
        """|coro|

        Creates a [text] channel in this guild.

        Parameters
        ----------
        name: str
            The name of the channel.

        Returns
        -------
        :class:`~.Channel`
        """
        c: ChannelPayload = await self._connection.api.guilds(self.id).channels.post(
            json={'name': name}
        )
        return Channel(self._connection, c)

    async def fetch_channel(self, id: Id) -> Channel:
        """|coro|

        Fetches a channel from this guild.

        Parameters
        ----------
        id: int
            The ID of the channel to fetch.

        Returns
        -------
        :class:`~.Channel`
        """
        id = sanitize_id(id)
        c = await self._connection.api.guilds(self.id).channels(id).get()
        return Channel(self._connection, c)

    async def edit(self, name: str) -> Guild:
        """|coro|

        Edits this guild.

        Parameters
        ----------
        name: str
            The new name of the guild.

        Returns
        -------
        :class:`~.Guild`
        """
        payload = {name: name}
        guild = await self._connection.api.guilds(self.id).patch(json=payload)
        self._process_data(guild)
        return self

    async def delete(self) -> None:
        """|coro|

        Deletes this guild.
        """
        await self._connection.api.guilds(self.id).delete()

    def get_channel(self, id: Id) -> Optional[Channel]:
        """Tries to retrieve a :class:`.~Channel` object
        given it's ID from the internal cache.

        Parameters
        ----------
        id: int
            The channel ID.

        Returns
        -------
        :class:`~.Channel`
            The channel found, if any.
        None
            The channel was not found in the internal cache.
        """
        id = sanitize_id(id)
        return self._channels.get(id)

    def get_member(self, id: Id) -> Optional[Member]:
        """Tries to retrieve a :class:`.~Member` object
        given it's ID from the internal cache.

        Parameters
        ----------
        id: int
            The member ID.

        Returns
        -------
        :class:`~.Member`
            The member foumd, if any.
        None
            The member was not found in the internal cache.
        """
        id = sanitize_id(id)
        return self._members.get(id)

    @property
    def owner(self) -> Optional[Member]:
        """Member: The owner of this guild."""
        return self.get_member(self._owner_id)

    @property
    def owner_id(self, /) -> Optional[Snowflake]:
        """int: The ID of the owner of this guild."""
        return self._owner_id

    @property
    def name(self, /) -> Optional[str]:
        """str: The name of this guild."""
        return self._name

    @property
    def channels(self, /) -> List[Channel]:
        """List[:class:`~.Channel`]: A list of all [cached] channels in this guild."""
        return list(self._channels.values())

    @property
    def members(self, /) -> List[Member]:
        """List[:class:`~.Member`]: A list of all [cached] members in this guild."""
        return list(self._members.values())

    def __repr__(self, /) -> str:
        return f'<Guild id={self.id} name={self.name!r}> owner_id={self.owner_id}'

    def __str__(self, /) -> str:
        return self.name or self.__repr__()



from __future__ import annotations

import asyncio
from asyncio import AbstractEventLoop
from collections import deque
from typing import TYPE_CHECKING, Any, Coroutine, Dict, Optional, Union

from ferris.types.base import Snowflake

from .http import APIRouter, HTTPClient
from .utils import find

if TYPE_CHECKING:
    from .channel import Channel
    from .guild import Guild
    from .message import Message
    from .user import User


__all__ = ('Connection',)


class Connection:
    def __init__(
        self, loop: AbstractEventLoop, dispatch: Coroutine, /, **options
    ) -> None:
        self.loop: AbstractEventLoop = loop
        self._user: Optional[User] = None

        self.dispatch: Coroutine = dispatch

        self._is_ready: asyncio.Future = loop.create_future()

        self._http: Union[HTTPClient, Any] = None
        self.__token: Optional[str] = None

        self._max_messages_count: int = options.get("max_messages_count", 1000)

        self.clear_store()

    @property
    def token(self) -> Optional[str]:
        return self.__token

    @property
    def api(self) -> Union[APIRouter, Any]:
        return self._http.api if self._http else None

    @property
    def user(self) -> Optional[User]:
        return self._user

    def _store_token(self, token: str, /) -> None:
        self.__token = token

    def _initialize_http(self, token: str, /) -> None:
        self._http = HTTPClient(token)
        self._store_token(token)

    async def _initialize_http_with_email(self, email: str, password: str, /) -> None:
        self._http = await HTTPClient.from_email_and_password(email, password)
        self._store_token(self._http.token)

    def clear_store(self, /) -> None:
        self._users: Dict[int, User] = {}
        self._guilds: Dict[int, Guild] = {}
        self._channels: Dict[int, Channel] = {}

        self._messages: deque = deque(maxlen=self._max_messages_count)

    def deref_user(self, id: int, /) -> None:
        self._users.pop(id, None)

    def deref_channel(self, id: int, /) -> None:
        self._channels.pop(id, None)

    def store_message(self, message: Message, /) -> None:
        self._messages.append(message)

    def remove_message(self, id: Snowflake, /) -> None:
        m = self.get_message(id)
        if not m:
            return
        self._messages.remove(m)

    def store_user(self, user: User, /) -> None:
        self._users[user.id] = user

    def store_guild(self, guild: Guild, /) -> None:
        self._guilds[guild.id] = guild

    def store_channel(self, channel: Channel, /) -> None:
        self._channels[channel.id] = channel

    def get_message(self, id: int, /) -> Optional[Message]:
        return find(lambda m: m.id == id, self._messages)

    def get_user(self, id: int, /) -> Optional[User]:
        return self._users.get(id)

    def get_guild(self, id: int, /) -> Optional[Guild]:
        return self._guilds.get(id)

    def get_channel(self, id: int, /) -> Optional[Channel]:
        return self._channels.get(id)



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, Union

from .base import BaseObject


if TYPE_CHECKING:
    from .connection import Connection
    from .types import Data, Snowflake
    from .types.message import MessagePayload
    from .channel import Channel
    from .user import User
    from .member import Member

__all__ = ('Message',)


class Message(BaseObject):
    """Represents a message from FerrisChat."""

    __slots__ = ('_connection', '_content', '_channel_id', '_author_id')

    def __init__(
        self, connection: Connection, data: Optional[MessagePayload], /
    ) -> None:
        self._connection: Connection = connection
        self._process_data(data)

    def _process_data(self, data: Optional[MessagePayload], /) -> None:
        if not data:
            return

        self._store_snowflake(data.get('id'))

        self._content: Optional[str] = data.get('content')
        self._channel_id: Optional[Snowflake] = data.get('channel_id')

        self._author_id: Optional[Snowflake] = data.get('author_id')

    async def edit(self, content: str) -> Message:
        """|coro|

        Edits this message.

        Parameters
        ----------
        content: str
            The new content for this message.

        Returns
        -------
        Message
            The edited message.
        """
        payload = {'content': content}
        m = (
            await self._connection.api.channels(self.channel_id)
            .messages(self.id)
            .patch(payload)
        )
        self._process_data(m)
        return self

    async def delete(self) -> None:
        """|coro|

        Deletes this message.
        """
        await self._connection.api.channels(self.channel_id).messages(self.id).delete()

    @property
    def author(self, /) -> Optional[Union[Member, User]]:
        if self.channel and self.channel.guild:
            return self.channel.guild.get_member(
                self.author_id
            ) or self._connection.get_user(self.author_id)
        return self._connection.get_user(self.author_id)

    @property
    def channel(self, /) -> Optional[Channel]:
        """Channel: The channel that this message was sent in"""
        return self._connection.get_channel(self.channel_id)

    @property
    def author_id(self, /) -> Optional[Snowflake]:
        """int: Returns the author ID of this message."""
        return self._author_id

    @property
    def content(self, /) -> Optional[str]:
        """str: The content of this message."""
        return self._content

    @property
    def channel_id(self, /) -> Optional[Snowflake]:
        """int: The ID of the channel this message was sent in."""
        return self._channel_id

    def __repr__(self, /) -> str:
        return f'<Message id={self.id} author_id={self.author_id} channel_id={self.channel_id}>'



from typing import Union

from .auth import *
from .base import *
from .channel import *
from .guild import *
from .member import *
from .message import *
from .user import *
from .ws import *


Data = Union[
    AuthResponse,
    ChannelPayload,
    GuildPayload,
    MemberPayload,
    MessagePayload,
    UserPayload,
]



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Optional, Union

from .utils import FERRIS_EPOCH


__all__ = ('Invite',)

if TYPE_CHECKING:
    from .connection import Connection
    from .guild import Guild
    from .member import Member
    from .types import Snowflake
    from .types.invite import InvitePayload
    from .user import User


class Invite:
    """
    Represents a invite to join guild.
    """

    __slots__ = (
        '_code',
        '_owner_id',
        '_guild_id',
        '_created_at',
        '_uses',
        '_max_uses',
        '_max_age',
        '_connection',
    )

    def __init__(self, connection: Connection, payload: Optional[InvitePayload]):
        self._connection = connection
        self._process_data(payload)

    def _process_data(self, data: Optional[InvitePayload], /) -> None:
        if not data:
            return
        self._code: str = data.get('code')
        self._owner_id: Snowflake = data.get('owner_id')
        self._guild_id: Snowflake = data.get('guild_id')
        self._created_at: datetime = datetime.fromtimestamp(
            (data.get('created_at', 0) + FERRIS_EPOCH)
        )
        self._uses: int = data.get('uses')
        self._max_uses: int = data.get('max_uses')
        self._max_age: int = data.get('max_age')

    @property
    def owner(self, /) -> Optional[Union[Member, User]]:
        """Optional[Union[Member, User]]: The owner of this invite."""
        if self.guild:
            return self.guild.get_member(self.owner_id) or self._connection.get_user(
                self.owner_id
            )
        return self._connection.get_user(self.owner_id)

    @property
    def guild(self, /) -> Optional[Guild]:
        """Optional[Guild]: The guild of this invite."""
        return self._connection.get_guild(self.guild_id)

    @property
    def code(self) -> str:
        """str: The invite code."""
        return self._code

    @property
    def guild_id(self) -> Snowflake:
        """Snowflake: The guild id this invite is for."""
        return self._guild_id

    @property
    def created_at(self) -> datetime:
        """datetime: The time this invite was created."""
        return self._created_at

    @property
    def uses(self) -> int:
        """int: The number of times this invite has been used."""
        return self._uses

    @property
    def max_uses(self) -> int:
        """int: The maximum number of times this invite can be used."""
        return self._max_uses

    @property
    def max_age(self) -> int:
        """int: The maximum age of this invite."""
        return self._max_age
    
    def __repr__(self) -> str:
        return f'<Invite code={self.code!r} guild={self.guild_id} uses={self.uses} max_uses={self.max_uses} max_age={self.max_age}>'



from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Coroutine

from .channel import Channel
from .guild import Guild
from .member import Member
from .message import Message
from .user import User
from .invite import Invite

if TYPE_CHECKING:
    from .connection import Connection

log = logging.getLogger(__name__)


class _BaseEventHandler:
    def __init__(self, connection: Connection) -> None:
        self.connection: Connection = connection
        self.dispatch: Coroutine = connection.dispatch
        self._is_ready: asyncio.Future = connection._is_ready

    def handle(self, _data: dict):
        event = _data.get('c')
        data = _data.get('d')
        log.debug(f"Handling event: {event} Data: {data}")

        self.dispatch('socket_receive', event, data)

        try:
            asyncio.create_task(getattr(self, event)(data))
        except AttributeError:
            log.error(f'Received unkwown event: {event}')


class EventHandler(_BaseEventHandler):
    async def IdentifyAccepted(self, data):
        self.dispatch('identify_accepted')
        u = User(self.connection, data.get('user', {}))

        self.connection._user = u

        self._is_ready.set_result(None)
        self.dispatch('ready')

    async def MessageCreate(self, data):
        m = Message(self.connection, data.get('message'))
        self.dispatch('message', m)
        self.connection.store_message(m)

    async def MessageUpdate(self, data):
        old = Message(self.connection, data.get('old'))

        if new := self.connection.get_message(old.id):
            new._process_data(data.get('new'))
        else:
            new = Message(self.connection, data.get('new'))

        self.dispatch('message_update', old, new)

    async def MessageDelete(self, data):
        if m := self.connection.get_message(data.get('id')):
            self.connection._messages.remove(m)
        else:
            m = Message(self.connection, data.get('message'))
            self.connection.remove_message(m.id)
        self.dispatch('message_delete', m)

    async def ChannelCreate(self, data):
        if c := self.connection.get_channel(data.get('id')):
            c._process_data(data.get('channel'))
        else:
            c = Channel(self.connection, data.get('channel'))
            self.connection.store_channel(c)
        self.dispatch('channel_create', c)

    async def ChannelUpdate(self, data):
        old = Channel(self.connection, data.get('old'))
        if new := self.connection.get_channel(old.id):
            new._process_data(data.get('new'))
        else:
            new = Channel(self.connection, data.get('new'))
            self.connection.store_channel(new)

        self.dispatch('channel_update', old, new)

    async def ChannelDelete(self, data):
        c = Channel(self.connection, data.get('channel'))
        self.dispatch('channel_delete', c)

        self.connection._channels.pop(c.id, None)
        
        self.connection.get_guild(c.guild_id)._channels.pop(c.id, None)

    async def MemberCreate(self, data):
        guild: Guild = self.connection.get_guild(data.get('guild_id'))
        if not guild:
            return

        if member := guild._members.get(data.get('user_id')):
            member._process_data(data.get('member'))
        else:
            member = Member(guild, data.get('member'))
            guild._members[member.id] = member
        self.dispatch('member_create', member)

    async def MemberUpdate(self, data):
        guild: Guild = self.connection.get_guild(data.get('guild_id'))
        if member := guild._members.get(data.get('user_id')):
            member._process_data(data.get('member'))
        else:
            member = Member(guild, data.get('user'))
            guild._members[member.id] = member
        self.dispatch('member_update', member)

    async def MemberDelete(self, data):
        guild: Guild = self.connection.get_guild(data.get('guild_id'))
        if member := guild._members.get(data.get('user_id')):
            guild._members.pop(member.id, None)
        else:
            member = Member(guild, data.get('member'))
        self.dispatch('member_delete', member)
    
    async def GuildCreate(self, data):
        g = Guild(self.connection, data.get('guild'))

        self.connection.store_guild(g)

        self.dispatch('guild_create', g)
    
    async def GuildUpdate(self, data):
        old = Guild(self.connection, data.get('old'))
        if new := self.connection.get_guild(old.id):
            new._process_data(data.get('new'))
        else:
            new = Guild(self.connection, data.get('new'))
            self.connection.store_guild(new)

        self.dispatch('guild_update', old, new)
    
    async def GuildDelete(self, data):
        g = Guild(self.connection, data.get('guild'))
        self.connection._guilds.pop(g.id, None)

        self.dispatch('guild_delete', g)
    
    async def InviteCreate(self, data):
        invite = Invite(self.connection, data.get('invite'))

        self.dispatch('invite_create', invite)
    
    async def InviteDelete(self, data):
        invite = Invite(self.connection, data.get('invite'))

        self.dispatch('invite_delete', invite)



from typing import TypedDict

from .base import Snowflake


__all__ = ('MessagePayload',)


class MessagePayload(TypedDict):
    id: Snowflake
    content: str
    channel_id: Snowflake
    author_id: Snowflake



from __future__ import annotations

from typing import (
    Any,
    Awaitable,
    Generic,
    List,
    Optional,
    TYPE_CHECKING,
    TypeVar,
    Union,
)

R = TypeVar('R')

if TYPE_CHECKING:
    from typing_extensions import Concatenate, ParamSpec

    from ferris.channel import Channel
    from ferris.guild import Guild
    from ferris.message import Message
    from ferris.member import Member
    from ferris.user import User

    P = ParamSpec('P')
    CommandCallbackT = Callable[Concatenate['Context', P], Awaitable[R]]
else:
    P = TypeVar('P')


class Command(Generic[P, R]):
    """Represents a command.

    Attributes
    ----------
    name: str
        The name of this command.
    aliases: List[str]
        A list of aliases for this command.
    callback: Callable[Concatenate[:class:`.Context`, P], Awaitable[R]]
        The command callback for this command.
    """

    def __init__(
        self, name: str, aliases: List[str], callback: CommandCallbackT
    ) -> None:
        self.name: str = name
        self.aliases: List[str] = aliases
        self.callback: CommandCallbackT = callback
        self.on_error: Optional[Callable[[Context, Exception], Any]] = None

    def error(self, func: Callable[[Context, Exception], Any]) -> None:
        """Adds an error handler to this command.

        Example
        -------
        .. code:: python3

            @bot.command()
            async def raise_error(ctx: Context) -> None:
                int('a')  # Will raise ValueError

            @raise_error.error
            async def error_handler(ctx: Context, exc: Exception) -> None:
                if isinstance(exc, ValueError):
                    await ctx.send('Got ValueError!')
        """
        self.on_error = func

    async def invoke(self, ctx: Context, *args: P.args, **kwargs: P.kwargs) -> None:
        """Invokes this command.

        Parameters
        ----------
        ctx: :class:`.Context`
            The context to invoke the command with.
        *args: P.args
            The positional arguments to pass into the command callback.
        **kwargs: P.kwargs
            The keyword arguments to pass into the command callback.
        """
        try:
            await self.callback(ctx, *args, **kwargs)
            # dispatch `on_command` here...
        except Exception as exc:
            if self.on_error:
                try:
                    await self.on_error(ctx, exc)
                except Exception as new_exc:
                    exc = new_exc
                else:
                    return
            # dispatch `on_command_error` here...
        else:
            # dispatch `on_command_success` here...
            ...
        finally:
            # dispatch `on_command_complete` here...
            ...


class Context:
    """Represents the context for a command.

    Attributes
    ----------
    command: Optional[:class:`.Command`]
        The command invoked. May be ``None``.
    """

    def __init__(self, *, message: Message) -> None:
        self._message: Message = message
        self.command: Command = None

    @property
    def message(self) -> Message:
        """:class:`~.Message`: The invocation message for this context."""
        return self._message

    @property
    def author(self) -> Union[User, Member]:
        """Union[:class:`~.User`, :class:`~.Member`]: The author of this context."""
        return self._message.author

    @property
    def channel(self) -> Channel:
        """:class:`~.Channel`: The channel of this context."""
        return self._message.channel

    @property
    def guild(self) -> Guild:
        """:class:`~.Guild`: The guild of this context."""
        return self._message.guild
